#
# pip install unstructured[all-docs]
# libmagic
#
# partition_object supports via unstructured library:
#
# "csv", "doc", "docx", "epub", "image", "md", "msg", "odt", "org",
# "pdf", "ppt", "pptx", "rtf", "rst", "tsv", "xlsx"

from transformers import pipeline

from dvcx.lib.dataset import Dataset
from dvcx.lib.unstructured import PartitionObject
from dvcx.query import C
from dvcx.sql.types import String

device = "cpu"
model = "pszemraj/led-large-book-summary"
source = "gcs://dvcx-datalakes/NLP/infobooks/"


def cleanse(text):
    separator = "Listen to this story"
    head, sep, tail = text.partition(separator)
    return (head,)


def summarize(clean):
    helper = pipeline(model=model, device=device)
    summary = helper(clean, max_length=200)[0]["summary_text"]
    return (summary,)


ds = (
    Dataset(
        source,
        anon=True,
    )
    .filter(C.name.glob("*.pdf"))
    .limit(1)
    .map(PartitionObject(), parallel=False)
)

ds = ds.map(cleanse, output={"clean": String})
ds = ds.map(summarize, output={"summary": String})
results = ds.select("text", "summary").results()

for story in results:
    print("\n *********** the original: ********** ")
    print(story[0])

    print("\n *********** the summary: *********** ")
    print(story[1])
    print("\n")



# pip install torch
import torch

from dvcx.lib.hf_pipeline import ImageHelper, RawHelper, TextHelper
from dvcx.query import C, DatasetQuery

image_source = "gcs://dvcx-datalakes/dogs-and-cats/"
audio_source = "gcs://dvcx-datalakes/speech-emotion-recognition-dataset/"
text_source = "gcs://dvcx-datalakes/NLP/cnn/stories"

if torch.cuda.is_available():
    device = "cuda"
else:
    device = "cpu"


if __name__ == "__main__":
    print("** HuggingFace pipeline helper model zoo demo **")
    print("\nZero-shot object detection and classification:")
    results = (
        DatasetQuery(
            image_source,
            anon=True,
        )
        .filter(C.name.glob("*.jpg"))
        .limit(1)
        .add_signals(
            ImageHelper(
                model="google/owlv2-base-patch16",
                device=device,
                candidate_labels=["cat", "dog", "squirrel", "unknown"],
            ),
            parallel=False,
        )
        .select("source", "parent", "name", "model_output", "error")
        .results()
    )
    print(*results, sep="\n")

    print("\nNot-safe-for-work image detection:")
    results = (
        DatasetQuery(
            image_source,
            anon=True,
        )
        .filter(C.name.glob("*.jpg"))
        .limit(1)
        .add_signals(
            ImageHelper(
                model="Falconsai/nsfw_image_detection",
                device=device,
            ),
            parallel=False,
        )
        .select("source", "parent", "name", "model_output", "error")
        .results()
    )
    print(*results, sep="\n")

    print("\nAudio emotion classification:")
    results = (
        DatasetQuery(
            audio_source,
            anon=True,
        )
        .filter(C.name.glob("*.wav"))
        .limit(1)
        .add_signals(
            RawHelper(
                model="Krithika-p/my_awesome_emotions_model",
                device=device,
            ),
            parallel=False,
        )
        .select("source", "parent", "name", "model_output", "error")
        .results()
    )
    print(*results, sep="\n")
    print("\nLong text summarization:")
    results = (
        DatasetQuery(
            text_source,
            anon=True,
        )
        .filter(C.name.glob("*.story"))
        .limit(1)
        .add_signals(
            TextHelper(
                model="pszemraj/led-large-book-summary",
                device=device,
                max_length=150,
            ),
            parallel=False,
        )
        .select("source", "parent", "name", "model_output", "error")
        .results()
    )
    print(*results, sep="\n")



import json

from dvcx.query import Object, udf
from dvcx.sql.types import JSON, String

try:
    from PIL import (
        ExifTags,
        Image,
        IptcImagePlugin,
        TiffImagePlugin,
        UnidentifiedImageError,
    )
except ImportError as exc:
    raise ImportError(
        "Missing dependency Pillow for computer vision:\n"
        "To install run:\n\n"
        "  pip install 'dvcx[cv]'\n"
    ) from exc


def encode_image(raw):
    try:
        img = Image.open(raw)
    except UnidentifiedImageError:
        return None
    return img


@udf(
    params=(Object(encode_image),),  # Columns consumed by the UDF.
    output={
        "xmp": JSON,
        "exif": JSON,
        "iptc": JSON,
        "error": String,
    },  # Signals being returned by the UDF.
    method="image_description",
)
class GetMetadata:
    def cast(self, v):  # to JSON serializable types
        if isinstance(v, TiffImagePlugin.IFDRational):
            return float(v)
        if isinstance(v, tuple):
            return tuple(self.cast(t) for t in v)
        if isinstance(v, bytes):
            return v.decode(encoding="utf-8", errors="ignore")
        if isinstance(v, dict):
            for kk, vv in v.items():
                v[kk] = self.cast(vv)
            return v
        if isinstance(v, list):
            return [self.cast(kk) for kk in v]
        return v

    def image_description(self, img):
        (xmp, exif, iptc) = ({}, {}, {})
        if img is None:
            error = "Image format not understood"
            return ({}, {}, {}, error)
        error = ""
        xmp = img.getxmp()
        img_exif = img.getexif()
        img_iptc = IptcImagePlugin.getiptcinfo(img)

        if img_iptc:
            for k, v in img_iptc.items():
                iptc[str(k)] = self.cast(v)

        if img_exif:
            for k, v in img_exif.items():
                v = self.cast(v)
                if k in ExifTags.TAGS:
                    exif[ExifTags.TAGS[k]] = v
                if k in ExifTags.GPSTAGS:
                    exif[ExifTags.GPSTAGS[k]] = v

        return (
            json.dumps(xmp),
            json.dumps(exif),
            json.dumps(iptc),
            error,
        )



import inspect
import re
import warnings
from collections.abc import Sequence
from datetime import datetime
from types import GenericAlias
from typing import Any, ClassVar, Literal, Union, get_args, get_origin

import attrs
from pydantic import BaseModel
from typing_extensions import Literal as LiteralEx

from dvcx.query import C
from dvcx.query.udf import UDFOutputSpec
from dvcx.sql.types import (
    JSON,
    Array,
    Binary,
    Boolean,
    DateTime,
    Float,
    Int,
    NullType,
    SQLType,
    String,
)

TYPE_TO_DVCX = {
    int: Int,
    str: String,
    Literal: String,
    LiteralEx: String,
    float: Float,
    bool: Boolean,
    datetime: DateTime,
    bytes: Binary,
    list: Array(NullType),
    dict: JSON,
    Any: NullType,
    None: NullType,
}


# Disable Pydantic warning, see https://github.com/iterative/dvcx/issues/1285
warnings.filterwarnings(
    "ignore",
    message="Field name .* shadows an attribute in parent",
    category=UserWarning,
)


class Feature(BaseModel):
    """A base class for defining data classes that serve as inputs and outputs for
    DataFrame processing functions like `map()`, `generate()`, etc. Inherits from
    `pydantic`'s BaseModel, allowing for data validation and definition.
    """

    _expand_name: ClassVar[bool] = True
    _delimiter: ClassVar[str] = "__"
    _is_file: ClassVar[bool] = False

    def get_value(self, *args: Any, **kwargs: Any) -> Any:
        name = self.__class__.__name__
        raise NotImplementedError(f"value is not defined for feature class {name}")

    def _get_value_with_check(self, *args: Any, **kwargs: Any) -> Any:
        signature = inspect.signature(self.get_value)
        for i, (name, prm) in enumerate(signature.parameters.items()):
            if prm.default == inspect.Parameter.empty:
                if i < len(args):
                    continue
                if name not in kwargs:
                    raise ValueError(
                        f"unable to get value for class {self.__class__.__name__}"
                        f" due to a missing parameter {name} in get_value()"
                    )

        return self.get_value(*args, **kwargs)

    @classmethod
    def __pydantic_init_subclass__(cls):
        for name, field_info in cls.model_fields.items():
            attr_value = _resolve(cls, name, field_info, cls._prefix())
            setattr(cls, name, RestrictedAttribute(attr_value, cls, name))

    @classmethod
    def _prefix(cls) -> str:
        return cls._normalize(cls.__name__)

    @classmethod
    def _normalize(cls, name: str) -> str:
        if (
            cls._expand_name
            and cls._delimiter
            and cls._delimiter.lower() in name.lower()
        ):
            raise RuntimeError(
                f"variable '{name}' cannot be used because it contains {cls._delimiter}"
            )
        return Feature._to_snake_case(name)

    @staticmethod
    def _to_snake_case(name: str) -> str:
        """Convert a CamelCase name to snake_case."""
        s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
        return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()

    @staticmethod
    def _convert_type(typ):  # noqa: PLR0911
        if inspect.isclass(typ) and issubclass(typ, SQLType):
            return typ
        res = TYPE_TO_DVCX.get(typ)
        if res:
            return res

        orig = get_origin(typ)

        if orig in (Literal, LiteralEx):
            return String

        if orig == list:
            args = get_args(typ)
            if args is None or len(args) != 1:
                raise TypeError(f"Cannot resolve type '{typ}' for flattening features")
            next_type = Feature._convert_type(args[0])
            return Array(next_type)

        if orig == dict:
            return JSON

        args = get_args(typ)
        if orig == Union and len(args) == 2 and (type(None) in args):
            return Feature._convert_type(args[0])

        # Special case for list in JSON: Union[dict, list[dict]]
        if orig == Union and len(args) >= 2:
            args_no_nones = [arg for arg in args if arg != type(None)]
            if len(args_no_nones) == 2:
                args_no_dicts = [arg for arg in args_no_nones if arg != dict]
                if len(args_no_dicts) == 1 and get_origin(args_no_dicts[0]) == list:
                    arg = get_args(args_no_dicts[0])
                    if len(arg) == 1 and arg[0] == dict:
                        return JSON

        raise RuntimeError(f"Cannot recognize type {typ}")

    @staticmethod
    def _iter_fields(fields):
        for name, f_info in fields.items():
            yield name, f_info.annotation, f_info.is_required()

    @classmethod
    def _flatten_full_schema(cls, fields, name_path):
        for name, anno, is_required in cls._iter_fields(fields):
            name = cls._normalize(name)

            orig = get_origin(anno)
            if orig == list:
                anno = get_args(anno)
                if isinstance(anno, tuple):
                    anno = anno[0]
                is_list = True
            else:
                is_list = False

            expanded_name = name
            if cls._expand_name:
                lst = [cls._prefix(), *name_path, name]
                expanded_name = cls._delimiter.join(lst)

            if Feature._is_feature_class(anno):
                if is_list:
                    yield expanded_name, Array(JSON), is_required
                else:
                    yield from cls._flatten_full_schema(
                        anno.model_fields, [*name_path, name]
                    )
            else:
                typ = Feature._convert_type(anno)
                if is_list:
                    typ = Array(typ)
                yield expanded_name, typ, is_required

    @classmethod
    def _is_feature_class(cls, anno):
        return (
            inspect.isclass(anno)
            and not isinstance(anno, GenericAlias)
            and issubclass(anno, Feature)
        )

    @classmethod
    def _to_udf_spec(cls):
        full_schema = cls._flatten_full_schema(cls.model_fields, [])
        return [(name, typ) for name, typ, is_required in full_schema]

    @staticmethod
    def _features_to_udf_spec(fr_classes: Sequence[type["Feature"]]) -> UDFOutputSpec:
        return dict(
            item
            for b in fr_classes
            for item in b._to_udf_spec()  # type: ignore[attr-defined]
        )

    def _flatten_fields_values(self, fields, dump):
        for name, anno, _ in self._iter_fields(fields):
            value = dump[name]
            if (
                inspect.isclass(anno)
                and anno.__module__ != "builtins"
                and issubclass(anno, Feature)
            ):
                yield from self._flatten_fields_values(anno.model_fields, value)
            else:
                yield value

    def _flatten(self):
        return tuple(self._flatten_fields_values(self.model_fields, self.model_dump()))

    @staticmethod
    def _flatten_list(objs):
        return tuple(val for obj in objs for val in obj._flatten())

    @classmethod
    def _unflatten_with_path(cls, dump, path):
        res = {}
        for name, anno, _ in Feature._iter_fields(cls.model_fields):
            name_norm = cls._normalize(name)
            if cls._expand_name:
                curr_path = path + cls._delimiter + name_norm
            else:
                curr_path = name_norm

            if inspect.isclass(anno) and issubclass(anno, Feature):
                val = anno._unflatten_with_path(dump, curr_path)
                res[name] = val
            else:
                res[name] = dump[curr_path]
        return cls(**res)

    @classmethod
    def _unflatten(cls, dump):
        return cls._unflatten_with_path(dump, cls._prefix())


class RestrictedAttribute:
    """Descriptor implementing an attribute that can only be accessed through
    the defining class and not from subclasses or instances.

    Since it is a non-data descriptor, instance dicts have precedence over it.
    Cannot be used with slotted classes.
    """

    def __init__(self, value, cls=None, name=None):
        self.cls = cls
        self.value = value
        self.name = name

    def __get__(self, instance, owner):
        if owner is not self.cls:
            raise AttributeError(
                f"'{type(owner).__name__}' object has no attribute '{self.name}'"
            )
        if instance is not None:
            raise RuntimeError(
                f"Invalid attempt to access class attribute '{self.name}' through "
                f"'{type(owner).__name__}' instance"
            )
        return self.value

    def __set_name__(self, cls, name):
        self.cls = cls
        self.name = name


@attrs.define
class FeatureAttributeWrapper:
    cls: type[Feature]
    prefix: str

    def __getattr__(self, name):
        field_info = self.cls.model_fields.get(name)
        if field_info:
            return _resolve(self.cls, name, field_info, prefix=self.prefix)
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'"
        )


def _resolve(cls, name, field_info, prefix):
    """Resolve feature attributes so they can be used in select(), join()
    and similar functions.

    Users just use `MyClass.sub_attr1.sub_attr2.field` and it will return a DB column
    with a proper name (with default naming - `my_class__sub_attr1__sub_attr2__field`).
    """
    anno = field_info.annotation
    norm_name = cls._normalize(name)

    if not cls._is_feature_class(anno):
        try:
            anno_sql_class = cls._convert_type(anno)
        except RuntimeError:
            anno_sql_class = NullType
        if cls._expand_name:
            return C(cls._delimiter.join([prefix, norm_name]), anno_sql_class)
        return C(norm_name, anno_sql_class)

    if not cls._expand_name:
        return FeatureAttributeWrapper(anno, "")

    new_prefix_value = cls._delimiter.join([prefix, norm_name])
    return FeatureAttributeWrapper(anno, new_prefix_value)


class ShallowFeature(Feature):
    _expand_name: ClassVar[bool] = False



import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable
from functools import cached_property
from typing import TYPE_CHECKING, Optional

from sqlalchemy import Column, Integer, Table, Text

from dvcx.data_storage.serializer import Serializable

if TYPE_CHECKING:
    from sqlalchemy.schema import SchemaItem

    from dvcx.data_storage.db_engine import DatabaseEngine


logger = logging.getLogger("dvcx")


class AbstractIDGenerator(ABC, Serializable):
    """
    Abstract ID Generator class. This class is responsible for generating
    unique IDs for each prefix (e.g. S3 bucket or dataset).
    """

    @abstractmethod
    def clone(self) -> "AbstractIDGenerator":
        """Clones AbstractIDGenerator implementation."""

    def init(self) -> None:
        """Initialize ID generator."""

    def cleanup_for_tests(self):
        """Cleanup for tests."""

    @abstractmethod
    def init_id(self, uri: str) -> None:
        """Initializes the ID generator for the given URI with zero last_id."""

    @abstractmethod
    def get_next_ids(self, uri: str, count: int) -> range:
        """Returns a range of IDs for the given URI."""

    def get_next_id(self, uri: str) -> int:
        """Returns the next ID for the given URI."""
        return self.get_next_ids(uri, 1)[0]

    def delete_uri(self, uri: str):
        """Deletes the given URI."""
        self.delete_uris([uri])

    def delete_uris(self, uris: Iterable[str]):
        """Deletes the given URIs."""


class AbstractDBIDGenerator(AbstractIDGenerator):
    """
    Abstract ID Generator class, to be implemented by any Database Adapters
    for a specific database system. This class is responsible for generating
    unique IDs for each prefix (e.g. S3 bucket or dataset) and storing them
    in a database. It is also responsible for initializing the database
    and creating the necessary tables.
    """

    _db: "DatabaseEngine"
    _table_prefix: Optional[str] = None
    _skip_db_init: bool = False
    _base_table_name = "id_generator"

    def __init__(
        self,
        db: "DatabaseEngine",
        table_prefix: Optional[str] = None,
        skip_db_init: bool = False,
    ):
        self._db = db
        self._table_prefix = table_prefix
        self._skip_db_init = skip_db_init
        if db and not skip_db_init:
            self.init()

    @abstractmethod
    def clone(self) -> "AbstractDBIDGenerator":
        """Clones AbstractIDGenerator implementation."""

    @property
    def db(self) -> "DatabaseEngine":
        return self._db

    @property
    def _columns(self) -> list["SchemaItem"]:
        return [
            Column("uri", Text, primary_key=True, nullable=False),
            # This is the last id used (and starts at zero if no ids have been used)
            Column("last_id", Integer, nullable=False),
        ]

    @cached_property
    def _table(self) -> Table:
        table_name = self._base_table_name
        if self._table_prefix:
            table_name = f"{self._table_prefix}_{table_name}"
        return Table(table_name, self.db.metadata, *self._columns, extend_existing=True)

    def init(self) -> None:
        self.db.create_table(self._table, if_not_exists=True)

    def cleanup_for_tests(self):
        """Cleanup for tests."""
        self.db.drop_table(self._table, if_exists=True)

    @abstractmethod
    def init_id(self, uri: str) -> None:
        """Initializes the ID generator for the given URI with zero last_id."""

    @abstractmethod
    def get_next_ids(self, uri: str, count: int) -> range:
        """Returns a range of IDs for the given URI."""

    def delete_uris(self, uris: Iterable[str]):
        """Deletes the given URIs from the database."""
        self.db.execute(self._table.delete().where(self._table.c.uri.in_(uris)))



from argparse import SUPPRESS, Action, ArgumentError, _AppendAction


class BooleanOptionalAction(Action):
    """
    Creates --[no-]option style bool options.

    Defined here since it doesn't exist in argparse in Python 3.8.

    Copied from:
    https://github.com/python/cpython/blob/c33aaa9d559398bbf2b80e891bf3ae6a716e4b8c/Lib/argparse.py#L863-L901
    """

    def __init__(
        self,
        option_strings,
        dest,
        default=None,
        type=None,
        choices=None,
        required=False,
        help=None,
        metavar=None,
    ):
        _option_strings = []
        for option_string in option_strings:
            _option_strings.append(option_string)

            if option_string.startswith("--"):
                option_string = "--no-" + option_string[2:]
                _option_strings.append(option_string)

        if help is not None and default is not None and default is not SUPPRESS:
            help += " (default: %(default)s)"

        super().__init__(
            option_strings=_option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string in self.option_strings:
            setattr(namespace, self.dest, not option_string.startswith("--no-"))

    def format_usage(self):
        return " | ".join(self.option_strings)


class CommaSeparatedArgs(_AppendAction):  # pylint: disable=protected-access
    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest) or []
        items.extend(v for value in values.split(",") if (v := value.strip()))
        setattr(namespace, self.dest, list(dict.fromkeys(items)))


class KeyValueArgs(_AppendAction):  # pylint: disable=protected-access
    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest) or {}
        for raw_value in filter(bool, values):
            key, sep, value = raw_value.partition("=")
            if not key or not sep or value == "":
                raise ArgumentError(self, f"expected 'key=value', got {raw_value!r}")
            items[key.strip()] = value

        setattr(namespace, self.dest, items)



"""Manages progress bars."""

import logging
import os
import re
import sys
from threading import RLock
from typing import Any, ClassVar

from fsspec.callbacks import TqdmCallback
from tqdm import tqdm

logger = logging.getLogger(__name__)
tqdm.set_lock(RLock())


def env2bool(var, undefined=False):
    """
    undefined: return value if env var is unset
    """
    var = os.getenv(var, None)
    if var is None:
        return undefined
    return bool(re.search("1|y|yes|true", var, flags=re.I))


class Tqdm(tqdm):
    """
    maximum-compatibility tqdm-based progressbars
    """

    BAR_FMT_DEFAULT = (
        "{percentage:3.0f}% {desc}|{bar}|"
        "{postfix[info]}{n_fmt}/{total_fmt}"
        " [{elapsed}<{remaining}, {rate_fmt:>11}]"
    )
    # nested bars should have fixed bar widths to align nicely
    BAR_FMT_DEFAULT_NESTED = (
        "{percentage:3.0f}%|{bar:10}|{desc:{ncols_desc}.{ncols_desc}}"
        "{postfix[info]}{n_fmt}/{total_fmt}"
        " [{elapsed}<{remaining}, {rate_fmt:>11}]"
    )
    BAR_FMT_NOTOTAL = "{desc}{bar:b}|{postfix[info]}{n_fmt} [{elapsed}, {rate_fmt:>11}]"
    BYTES_DEFAULTS: ClassVar[dict[str, Any]] = {
        "unit": "B",
        "unit_scale": True,
        "unit_divisor": 1024,
        "miniters": 1,
    }

    def __init__(
        self,
        iterable=None,
        disable=None,
        level=logging.ERROR,
        desc=None,
        leave=False,
        bar_format=None,
        bytes=False,
        file=None,
        total=None,
        postfix=None,
        **kwargs,
    ):
        """
        bytes   : shortcut for
            `unit='B', unit_scale=True, unit_divisor=1024, miniters=1`
        desc  : persists after `close()`
        level  : effective logging level for determining `disable`;
            used only if `disable` is unspecified
        disable  : If (default: None) or False,
            will be determined by logging level.
            May be overridden to `True` due to non-TTY status.
            Skip override by specifying env var `DVC_IGNORE_ISATTY`.
        kwargs  : anything accepted by `tqdm.tqdm()`
        """
        kwargs = kwargs.copy()
        if bytes:
            kwargs = self.BYTES_DEFAULTS | kwargs
        else:
            kwargs.setdefault("unit_scale", total > 999 if total else True)
        if file is None:
            file = sys.stderr
        # auto-disable based on `logger.level`
        if not disable:
            disable = logger.getEffectiveLevel() > level
        # auto-disable based on TTY
        if (
            not disable
            and not env2bool("DVC_IGNORE_ISATTY")
            and hasattr(file, "isatty")
        ):
            disable = not file.isatty()
        super().__init__(
            iterable=iterable,
            disable=disable,
            leave=leave,
            desc=desc,
            bar_format="!",
            lock_args=(False,),
            total=total,
            **kwargs,
        )
        self.postfix = postfix or {"info": ""}
        if bar_format is None:
            if self.__len__():
                self.bar_format = (
                    self.BAR_FMT_DEFAULT_NESTED if self.pos else self.BAR_FMT_DEFAULT
                )
            else:
                self.bar_format = self.BAR_FMT_NOTOTAL
        else:
            self.bar_format = bar_format
        self.refresh()

    def close(self):
        self.postfix["info"] = ""
        # remove ETA (either unknown or zero); remove completed bar
        self.bar_format = self.bar_format.replace("<{remaining}", "").replace(
            "|{bar:10}|", " "
        )
        super().close()

    @property
    def format_dict(self):
        """inject `ncols_desc` to fill the display width (`ncols`)"""
        d = super().format_dict
        ncols = d["ncols"] or 80
        # assumes `bar_format` has max one of ("ncols_desc" & "ncols_info")

        meter = self.format_meter(  # type: ignore[call-arg]
            ncols_desc=1, ncols_info=1, **d
        )
        ncols_left = ncols - len(meter) + 1
        ncols_left = max(ncols_left, 0)
        if ncols_left:
            d["ncols_desc"] = d["ncols_info"] = ncols_left
        else:
            # work-around for zero-width description
            d["ncols_desc"] = d["ncols_info"] = 1
            d["prefix"] = ""
        return d


class CombinedDownloadCallback(TqdmCallback):
    def set_size(self, size):
        # This is a no-op to prevent fsspec's .get_file() from setting the combined
        # download size to the size of the current file.
        pass



import hashlib
import os
from functools import partial
from typing import TYPE_CHECKING, Optional

import attrs
from dvc_data.hashfile.db.local import LocalHashFileDB
from dvc_objects.fs.local import LocalFileSystem
from fsspec.callbacks import Callback, TqdmCallback

from .progress import Tqdm

if TYPE_CHECKING:
    from dvcx.client import Client
    from dvcx.storage import StorageURI

sha256 = partial(hashlib.sha256, usedforsecurity=False)


@attrs.frozen
class UniqueId:
    storage: "StorageURI"
    parent: str
    name: str
    etag: str
    size: int
    vtype: str
    location: Optional[str]

    @property
    def path(self) -> str:
        return f"{self.parent}/{self.name}" if self.parent else self.name

    def get_hash(self) -> str:
        etag = f"{self.vtype}{self.location}" if self.vtype else self.etag
        return sha256(
            f"{self.storage}/{self.parent}/{self.name}/{etag}".encode()
        ).hexdigest()


def try_scandir(path):
    try:
        with os.scandir(path) as it:
            yield from it
    except OSError:
        pass


class DVCXCache:
    def __init__(self, cache_dir: str, tmp_dir: str):
        self.odb = LocalHashFileDB(
            LocalFileSystem(),
            cache_dir,
            tmp_dir=tmp_dir,
        )

    @property
    def cache_dir(self):
        return self.odb.path

    @property
    def tmp_dir(self):
        return self.odb.tmp_dir

    def get_path(self, uid: UniqueId) -> Optional[str]:
        if self.contains(uid):
            return self.path_from_checksum(uid.get_hash())
        return None

    def contains(self, uid: UniqueId) -> bool:
        return self.odb.exists(uid.get_hash())

    def path_from_checksum(self, checksum: str) -> str:
        assert checksum
        return self.odb.oid_to_path(checksum)

    def remove(self, uid: UniqueId) -> None:
        self.odb.delete(uid.get_hash())

    async def download(
        self, uid: UniqueId, client: "Client", callback: Optional[Callback] = None
    ) -> None:
        from_path = f"{uid.storage}/{uid.path}"
        from dvc_objects.fs.utils import tmp_fname

        odb_fs = self.odb.fs
        tmp_info = odb_fs.join(self.odb.tmp_dir, tmp_fname())  # type: ignore[arg-type]
        size = uid.size
        if size < 0:
            size = await client.get_size(from_path)
        cb = callback or TqdmCallback(
            tqdm_kwargs={"desc": odb_fs.name(from_path), "bytes": True},
            tqdm_cls=Tqdm,
            size=size,
        )
        try:
            await client.get_file(from_path, tmp_info, callback=cb)
        finally:
            if not callback:
                cb.close()

        try:
            oid = uid.get_hash()
            self.odb.add(tmp_info, self.odb.fs, oid)
        finally:
            os.unlink(tmp_info)

    def store_data(self, uid: UniqueId, contents: bytes) -> None:
        checksum = uid.get_hash()
        dst = self.path_from_checksum(checksum)
        if not os.path.exists(dst):
            # Create the file only if it's not already in cache
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            with open(dst, mode="wb") as f:
                f.write(contents)

    def clear(self):
        """
        Completely clear the cache.
        """
        self.odb.clear()

    def get_total_size(self) -> int:
        total = 0
        for subdir in try_scandir(self.odb.path):
            for file in try_scandir(subdir):
                try:
                    total += file.stat().st_size
                except OSError:
                    pass
        return total



"""
SQL types.

This module provides SQL types to provide common features and interoperability
between different database backends which often have different typing systems.

See https://docs.sqlalchemy.org/en/20/core/custom_types.html#sqlalchemy.types.TypeDecorator.load_dialect_impl

For the corresponding python to db type conversion, it's often simpler and
more direct to use methods at the DBAPI rather than sqlalchemy. For example
for sqlite we can use `sqlite.register_converter`
( https://docs.python.org/3/library/sqlite3.html#sqlite3.register_converter )
"""

from datetime import datetime
from types import MappingProxyType
from typing import Any, Union

from sqlalchemy import TypeDecorator, types

_registry: dict[str, "TypeConverter"] = {}
registry = MappingProxyType(_registry)

_read_converter_registry: dict[str, "TypeReadConverter"] = {}
read_converter_registry = MappingProxyType(_read_converter_registry)

_type_defaults_registry: dict[str, "TypeDefaults"] = {}
type_defaults_registry = MappingProxyType(_type_defaults_registry)

NullType = types.NullType


def register_backend_types(dialect_name: str, type_cls):
    _registry[dialect_name] = type_cls


def register_type_read_converters(dialect_name: str, trc: "TypeReadConverter"):
    _read_converter_registry[dialect_name] = trc


def register_type_defaults(dialect_name: str, td: "TypeDefaults"):
    _type_defaults_registry[dialect_name] = td


def converter(dialect) -> "TypeConverter":
    name = dialect.name
    try:
        return registry[name]
    except KeyError:
        raise ValueError(
            f"No type converter registered for dialect: {dialect.name!r}"
        ) from None


def read_converter(dialect) -> "TypeReadConverter":
    name = dialect.name
    try:
        return read_converter_registry[name]
    except KeyError:
        raise ValueError(
            f"No read type converter registered for dialect: {dialect.name!r}"
        ) from None


def type_defaults(dialect) -> "TypeDefaults":
    name = dialect.name
    try:
        return type_defaults_registry[name]
    except KeyError:
        raise ValueError(f"No type defaults registered for dialect: {name!r}") from None


class SQLType(TypeDecorator):
    impl: type[types.TypeEngine[Any]] = types.TypeEngine
    cache_ok = True

    def to_dict(self) -> dict[str, Any]:
        return {"type": self.__class__.__name__}

    @classmethod
    def from_dict(cls, _: dict[str, Any]) -> Union[type["SQLType"], "SQLType"]:
        return cls


class String(SQLType):
    impl = types.String

    @property
    def python_type(self):
        return str

    def load_dialect_impl(self, dialect):
        return converter(dialect).string()

    @staticmethod
    def default_value(dialect):
        return type_defaults(dialect).string()

    def on_read_convert(self, value, dialect):
        return read_converter(dialect).string(value)


class Boolean(SQLType):
    impl = types.Boolean

    @property
    def python_type(self):
        return bool

    def load_dialect_impl(self, dialect):
        return converter(dialect).boolean()

    @staticmethod
    def default_value(dialect):
        return type_defaults(dialect).boolean()

    def on_read_convert(self, value, dialect):
        return read_converter(dialect).boolean(value)


class Int(SQLType):
    impl = types.INTEGER

    @property
    def python_type(self):
        return int

    def load_dialect_impl(self, dialect):
        return converter(dialect).int()

    @staticmethod
    def default_value(dialect):
        return type_defaults(dialect).int()

    def on_read_convert(self, value, dialect):
        return read_converter(dialect).int(value)


class Int32(Int):
    def load_dialect_impl(self, dialect):
        return converter(dialect).int32()

    @staticmethod
    def default_value(dialect):
        return type_defaults(dialect).int32()

    def on_read_convert(self, value, dialect):
        return read_converter(dialect).int32(value)


class Int64(Int):
    def load_dialect_impl(self, dialect):
        return converter(dialect).int64()

    @staticmethod
    def default_value(dialect):
        return type_defaults(dialect).int64()

    def on_read_convert(self, value, dialect):
        return read_converter(dialect).int64(value)


class UInt64(Int):
    def load_dialect_impl(self, dialect):
        return converter(dialect).uint64()

    @staticmethod
    def default_value(dialect):
        return type_defaults(dialect).uint64()

    def on_read_convert(self, value, dialect):
        return read_converter(dialect).uint64(value)


class Float(SQLType):
    impl = types.INTEGER

    @property
    def python_type(self):
        return float

    def load_dialect_impl(self, dialect):
        return converter(dialect).float()

    @staticmethod
    def default_value(dialect):
        return type_defaults(dialect).float()

    def on_read_convert(self, value, dialect):
        return read_converter(dialect).float(value)


class Float32(Float):
    def load_dialect_impl(self, dialect):
        return converter(dialect).float32()

    @staticmethod
    def default_value(dialect):
        return type_defaults(dialect).float32()

    def on_read_convert(self, value, dialect):
        return read_converter(dialect).float32(value)


class Float64(Float):
    def load_dialect_impl(self, dialect):
        return converter(dialect).float64()

    @staticmethod
    def default_value(dialect):
        return type_defaults(dialect).float64()

    def on_read_convert(self, value, dialect):
        return read_converter(dialect).float64(value)


class Array(SQLType):
    impl = types.ARRAY

    @property
    def python_type(self):
        return list

    def load_dialect_impl(self, dialect):
        return converter(dialect).array(self.item_type)

    def to_dict(self) -> dict[str, Any]:
        item_type_dict = (
            self.item_type.to_dict()
            if isinstance(self.item_type, SQLType)
            else self.item_type().to_dict()
        )
        return {
            "type": self.__class__.__name__,
            "item_type": item_type_dict,
        }

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Union[type["SQLType"], "SQLType"]:
        sub_t = NAME_TYPES_MAPPING[d["item_type"]["type"]].from_dict(  # type: ignore [attr-defined]
            d["item_type"]
        )
        return cls(sub_t)

    @staticmethod
    def default_value(dialect):
        return type_defaults(dialect).array()

    def on_read_convert(self, value, dialect):
        return read_converter(dialect).array(value, self.item_type, dialect)


class JSON(SQLType):
    impl = types.JSON

    @property
    def python_type(self):
        return dict

    def load_dialect_impl(self, dialect):
        return converter(dialect).json()

    @staticmethod
    def default_value(dialect):
        return type_defaults(dialect).json()

    def on_read_convert(self, value, dialect):
        return read_converter(dialect).json(value)


class DateTime(SQLType):
    impl = types.DATETIME

    @property
    def python_type(self):
        return datetime

    def load_dialect_impl(self, dialect):
        return converter(dialect).datetime()

    @staticmethod
    def default_value(dialect):
        return type_defaults(dialect).datetime()

    def on_read_convert(self, value, dialect):
        return read_converter(dialect).datetime(value)


class Binary(SQLType):
    impl = types.BINARY

    @property
    def python_type(self):
        return bytes

    def load_dialect_impl(self, dialect):
        return converter(dialect).binary()

    @staticmethod
    def default_value(dialect):
        return type_defaults(dialect).binary()

    def on_read_convert(self, value, dialect):
        return read_converter(dialect).binary(value)


class TypeReadConverter:
    def string(self, value):
        return value

    def boolean(self, value):
        return value

    def int(self, value):
        return value

    def int32(self, value):
        return value

    def int64(self, value):
        return value

    def uint64(self, value):
        return value

    def float(self, value):
        return value

    def float32(self, value):
        return value

    def float64(self, value):
        return value

    def array(self, value, item_type, dialect):
        if value is None or item_type is None:
            return value
        return [item_type.on_read_convert(x, dialect) for x in value]

    def json(self, value):
        return value

    def datetime(self, value):
        return value

    def uuid(self, value):
        return value

    def binary(self, value):
        return value


class TypeConverter:
    def string(self):
        return types.String()

    def boolean(self):
        return types.Boolean()

    def int(self):
        return types.Integer()

    def int32(self):
        return self.int()

    def int64(self):
        return self.int()

    def uint64(self):
        return self.int()

    def float(self):
        return types.Float()

    def float32(self):
        return self.float()

    def float64(self):
        return self.float()

    def array(self, item_type):
        return types.ARRAY(item_type)

    def json(self):
        return types.JSON()

    def datetime(self):
        return types.DATETIME()

    def binary(self):
        return types.BINARY()


class TypeDefaults:
    def string(self):
        return None

    def boolean(self):
        return None

    def int(self):
        return None

    def int32(self):
        return None

    def int64(self):
        return None

    def uint64(self):
        return None

    def float(self):
        return None

    def float32(self):
        return None

    def float64(self):
        return None

    def array(self):
        return None

    def json(self):
        return None

    def datetime(self):
        return None

    def uuid(self):
        return None

    def binary(self):
        return None


TYPES = [
    String,
    Boolean,
    Int,
    Int32,
    Int64,
    UInt64,
    Float,
    Float32,
    Float64,
    Array,
    JSON,
    DateTime,
    Binary,
]

NAME_TYPES_MAPPING = {t.__name__: t for t in TYPES}



import os

import pandas as pd

from dvcx.lib.claude import AggregateTextAnalytics, Summary
from dvcx.lib.dataset import Dataset
from dvcx.query.schema import C
from dvcx.sql.functions import path

SOURCE = "s3://ldb-public/remote/chatbots-public/"
model = "claude-3-opus-20240229"

PROMPT = """Consider the following dialogues between the 'user' and the 'bot' separated\
 by '===='. The 'user' is a human trying to find the best mobile plan. The 'bot' is a \
chatbot designed to query the user and offer the best solution. The dialog is \
successful if the 'bot' is able to gather the information and offer a plan, or inform \
the user that such plan does not exist. The dialog is not successful if the \
conversation ends early or the 'user' requests additional functions the 'bot' \
cannot perform. Read the dialogues and classify them into a fixed number of concise \
failure reasons covering most failure cases. Present output as JSON list of reason \
strings and nothing else.
"""

chats = Dataset(SOURCE, anon=True).filter(C.name.glob("*.txt"))

content = chats.limit(50).aggregate(
    AggregateTextAnalytics(
        prompt=PROMPT,
        model_name=model,
        messages=[
            {
                "role": "assistant",
                "content": "Here is the JSON object in curly brackets:",
            }
        ],
        api_key=os.getenv("ANTHROPIC_API_KEY"),
    ),
    partition_by=path.file_ext(C.name),
)

with pd.option_context("display.max_columns", None):
    df = content.select(Summary).to_pandas()
    print(df)

    errors = df[df.summary__error != ""]
    if len(errors) > 0:
        print("Errors: ", len(errors))



import hashlib
import json
import tarfile
from collections.abc import Sequence
from typing import Any, ClassVar, Optional, Union, get_args, get_origin

from pydantic import Field

from dvcx.lib.feature import Feature
from dvcx.lib.feature_udf import FeatureGenerator
from dvcx.lib.file import File, FileInfo, TarVFile
from dvcx.lib.utils import DvcxError


class WDSError(DvcxError):
    def __init__(self, tar_stream, message: str):
        super().__init__(f"WebDataset error '{tar_stream.get_full_name()}': {message}")


class CoreFileDuplicationError(WDSError):
    def __init__(self, tar_stream, file1: str, file2: str):
        super().__init__(
            tar_stream, f"duplication of files with core extensions: {file1}, {file2}"
        )


class CoreFileNotFoundError(WDSError):
    def __init__(self, tar_stream, extensions, stem):
        super().__init__(
            tar_stream,
            f"no files with the extensions '{','.join(extensions)}'"
            f" were found for file stem {stem}",
        )


class UnknownFileExtensionError(WDSError):
    def __init__(self, tar_stream, name, ext):
        super().__init__(tar_stream, f"unknown extension '{ext}' for file '{name}'")


class WDSBasic(Feature):
    pass


class WDSAllFile(WDSBasic):
    txt: Optional[str] = Field(default=None)
    text: Optional[str] = Field(default=None)
    cap: Optional[str] = Field(default=None)
    transcript: Optional[str] = Field(default=None)
    cls: Optional[int] = Field(default=None)
    cls2: Optional[int] = Field(default=None)
    index: Optional[int] = Field(default=None)
    inx: Optional[int] = Field(default=None)
    id: Optional[int] = Field(default=None)
    json: Optional[dict] = Field(default=None)  # type: ignore[assignment]
    jsn: Optional[dict] = Field(default=None)

    pyd: Optional[bytes] = Field(default=None)
    pickle: Optional[bytes] = Field(default=None)
    pth: Optional[bytes] = Field(default=None)
    ten: Optional[bytes] = Field(default=None)
    tb: Optional[bytes] = Field(default=None)
    mp: Optional[bytes] = Field(default=None)
    msg: Optional[bytes] = Field(default=None)
    npy: Optional[bytes] = Field(default=None)
    npz: Optional[bytes] = Field(default=None)
    cbor: Optional[bytes] = Field(default=None)


class WDSReadableSubclass(Feature):
    @staticmethod
    def _reader(builder, item: tarfile.TarInfo) -> "WDSReadableSubclass":
        raise NotImplementedError


class BuilderState:
    def __init__(self):
        self.stem = None
        self.core_file = None
        self.data = {}


class Builder:
    DEFAULT_TYPES_READERS: ClassVar[dict[type, Any]] = {
        str: lambda bld, item: bld.read_text(item),
        int: lambda bld, item: int(bld.read_text(item)),
        float: lambda bld, item: float(bld.read_text(item)),
        bytes: lambda bld, item: bld.read(item),
        dict: lambda bld, item: json.loads(bld.read_text(item)),
    }

    def __init__(
        self,
        tar_stream: File,
        core_extensions: list[str],
        wds_class: type[WDSBasic],
        tar,
        encoding="utf-8",
    ):
        self._core_extensions = core_extensions
        self._tar_stream = tar_stream
        self._wds_class = wds_class
        self._tar = tar
        self._encoding = encoding
        self.state = BuilderState()

    def read(self, item):
        return self._tar.extractfile(item).read()

    def read_text(self, item):
        return self._tar.extractfile(item).read().decode(self._encoding)

    def add(self, file: tarfile.TarInfo):
        fstream = File(name=file.name)
        ext = fstream.get_file_ext()
        stem = fstream.get_file_stem()

        if self.state.stem is not None and self.state.stem != stem:
            raise StopIteration

        if self.state.stem is None:
            self.state.stem = stem

        if ext in self._core_extensions:
            if self.state.core_file is not None:
                raise CoreFileDuplicationError(
                    self._tar_stream, file.name, self.state.core_file.name
                )
            self.state.core_file = file
        elif ext in self.state.data:
            raise WDSError(
                self._tar_stream,
                f"file with extension '.{ext}' already exists in the archive",
            )
        else:
            type_ = self._get_type(ext)
            if type_ is None:
                raise UnknownFileExtensionError(self._tar_stream, fstream.name, ext)

            if issubclass(type_, WDSReadableSubclass):
                reader = type_._reader
            else:
                reader = self.DEFAULT_TYPES_READERS.get(type_, None)

            if reader is None:
                raise WDSError(
                    self._tar_stream,
                    f"unable to find a reader for type {type_}, extension .{ext}",
                )
            self.state.data[ext] = reader(self, file)

    def produce(self):
        if self.state.core_file is None:
            raise CoreFileNotFoundError(
                self._tar_stream, self._core_extensions, self.state.stem
            )

        wds = self._wds_class(**self.state.data)

        new_parent = f"{self._tar_stream.get_full_name()}"
        core_file = self.state.core_file
        etag_string = "-".join(
            [self._tar_stream.etag, core_file.name, str(core_file.mtime)]
        )
        etag = hashlib.md5(etag_string.encode(), usedforsecurity=False).hexdigest()

        f_info = FileInfo(
            name=core_file.name,
            source=self._tar_stream.source,
            parent=new_parent,
            size=core_file.size,
            etag=etag,
            vtype="",
            location=[
                {
                    "vtype": TarVFile.get_vtype(),
                    "parent": self._tar_stream.model_dump(),
                    "size": core_file.size,
                    "offset": core_file.offset_data,
                }
            ],
        )

        self.state = BuilderState()

        return f_info, wds

    def _get_type(self, ext):
        field = self._wds_class.model_fields.get(ext, None)
        if field is None:
            return

        anno = field.annotation
        if get_origin(anno) == Union:
            args = get_args(anno)
            anno = args[0]

        return anno


class TarStream(File):
    @staticmethod
    def to_text(data):
        return data.decode("utf-8")

    _DATA_CONVERTERS: ClassVar[dict[type, Any]] = {
        str: lambda data: TarStream.to_text(data),
        int: lambda data: int(TarStream.to_text(data)),
        float: lambda data: float(TarStream.to_text(data)),
        bytes: lambda data: data,
        dict: lambda data: json.loads(TarStream.to_text(data)),
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._tar = None

    def open(self):
        self._tar = tarfile.open(fileobj=super().open())
        return self

    def getmembers(self) -> list[tarfile.TarInfo]:
        return self._tar.getmembers()

    def read_member(self, member: tarfile.TarInfo, type):
        fd = self._tar.extractfile(member)
        data = fd.read()
        converter = self._DATA_CONVERTERS.get(type, None)
        if not converter:
            raise ValueError("")
        return converter(data)


class WebDataset(FeatureGenerator):
    def __init__(
        self,
        core_extensions: Sequence[str] = ("jpg", "png"),
        spec=WDSAllFile,
        encoding="utf-8",
    ):
        self.core_extensions = core_extensions
        self.encoding = encoding
        self.spec = spec

        super().__init__(File, [FileInfo, spec])

    def process(self, stream: File):
        with stream.open() as fd:
            with tarfile.open(fileobj=fd) as tar:
                yield from self.get_tar_groups(
                    stream, tar, self.core_extensions, self.spec, self.encoding
                )

    @staticmethod
    def get_tar_groups(stream, tar, core_extensions, spec, encoding="utf-8"):
        builder = Builder(stream, core_extensions, spec, tar, encoding)

        for item in tar.getmembers():
            if not item.isfile():
                continue

            try:
                builder.add(item)
            except StopIteration:
                yield builder.produce()
                builder.add(item)
        if builder.state.stem is not None:
            yield builder.produce()



from dvcx.sql import literal, select
from dvcx.sql.functions import array, string


def test_length(warehouse):
    query = select(
        array.length(["abc", "def", "g", "hi"]),
        array.length([3.0, 5.0, 1.0, 6.0, 1.0]),
        array.length([[1, 2, 3], [4, 5, 6]]),
    )
    result = tuple(warehouse.db.execute(query))
    assert result == ((4, 5, 2),)


def test_length_on_split(warehouse):
    query = select(
        array.length(string.split(literal("abc/def/g/hi"), literal("/"))),
    )
    result = tuple(warehouse.db.execute(query))
    assert result == ((4,),)



import io
from typing import IO

from fsspec.callbacks import DEFAULT_CALLBACK, Callback


class FileWrapper(io.RawIOBase):
    """Instrumented wrapper around an existing file object.

    It wraps the file's read() method to update the callback with the number of
    bytes read.

    It assumes exclusive access to the underlying file object and closes it when it
    gets closed itself.
    """

    def __init__(self, fileobj, callback: Callback = DEFAULT_CALLBACK):
        self.fileobj = fileobj
        self.callback = callback

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False

    def seekable(self) -> bool:
        return self.fileobj.seekable()

    def tell(self):
        """Return the current file position."""
        return self.fileobj.tell()

    def seek(self, position, whence=io.SEEK_SET):
        """Seek to a position in the file."""
        return self.fileobj.seek(position, whence)

    def readinto(self, b) -> int:
        res = self.fileobj.readinto(b)
        self.callback.relative_update(res)
        return res

    def close(self):
        self.fileobj.close()
        super().close()


class FileSlice(io.RawIOBase):
    """A thin wrapper around an existing file object that provides a part of its data
    as an individual file object.

    It assumes exclusive access to the underlying file object and closes it when it
    gets closed itself.
    """

    def __init__(self, fileobj: IO[bytes], offset: int, size: int, name: str):
        self.fileobj = fileobj
        self.offset = offset
        self.size = size
        self.position = 0
        self.name = name

    def readable(self):
        return True

    def writable(self):
        return False

    def seekable(self):
        return self.fileobj.seekable()

    def tell(self):
        """Return the current file position."""
        return self.position

    def seek(self, position, whence=io.SEEK_SET):
        """Seek to a position in the file."""
        if whence == io.SEEK_SET:
            self.position = min(max(position, 0), self.size)
        elif whence == io.SEEK_CUR:
            if position < 0:
                self.position = max(self.position + position, 0)
            else:
                self.position = min(self.position + position, self.size)
        elif whence == io.SEEK_END:
            self.position = max(min(self.size + position, self.size), 0)
        else:
            raise ValueError("Invalid argument")
        return self.position

    def readinto(self, b):
        max_size = self.size - self.position
        if max_size <= 0:
            return 0
        self.fileobj.seek(self.offset + self.position)
        if len(b) > max_size:
            b = memoryview(b)[:max_size]
        res = self.fileobj.readinto(b)
        if res != len(b):
            raise RuntimeError("unexpected end of data")
        self.position += res
        return res

    def close(self):
        self.fileobj.close()
        super().close()



from sqlalchemy import Column
from sqlalchemy.dialects.sqlite import dialect as sqlite_dialect
from sqlalchemy.schema import CreateTable

from dvcx.data_storage.schema import DatasetRow
from dvcx.sql.types import JSON, Array, Binary, Float, Float32, Float64, Int


def test_dataset_table_compilation():
    table = DatasetRow.new_table(
        "ds-1",
        custom_columns=[
            Column("score", Float, nullable=False),
            Column("meta_info", JSON),
        ],
    )
    result = CreateTable(table, if_not_exists=True).compile(dialect=sqlite_dialect())

    assert result.string == (
        "\n"
        'CREATE TABLE IF NOT EXISTS "ds-1" (\n'
        "\tid INTEGER NOT NULL, \n"
        "\tvtype VARCHAR NOT NULL, \n"
        "\tdir_type INTEGER, \n"
        "\tparent VARCHAR, \n"
        "\tname VARCHAR NOT NULL, \n"
        "\tetag VARCHAR, \n"
        "\tversion VARCHAR, \n"
        "\tis_latest BOOLEAN, \n"
        "\tlast_modified DATETIME, \n"
        "\tsize INTEGER NOT NULL, \n"
        "\towner_name VARCHAR, \n"
        "\towner_id VARCHAR, \n"
        "\trandom INTEGER NOT NULL, \n"
        "\tlocation JSON, \n"
        "\tsource VARCHAR NOT NULL, \n"
        "\tscore FLOAT NOT NULL, \n"
        "\tmeta_info JSON, \n"
        "\tPRIMARY KEY (id)\n"
        ")\n"
        "\n"
    )


def test_custom_column_types_serialization(dataset_record):
    dataset_record.custom_column_types = {"int_col": Int}
    assert dataset_record.custom_column_types_serialized == {"int_col": {"type": "Int"}}

    dataset_record.custom_column_types = {
        "binary_col": Binary,
        "float_32_col": Float32,
    }
    assert dataset_record.custom_column_types_serialized == {
        "binary_col": {"type": "Binary"},
        "float_32_col": {"type": "Float32"},
    }

    dataset_record.custom_column_types = {"nested_col": Array(Array(Float64))}
    assert dataset_record.custom_column_types_serialized == {
        "nested_col": {
            "type": "Array",
            "item_type": {"type": "Array", "item_type": {"type": "Float64"}},
        }
    }



from datetime import datetime
from decimal import Decimal
from typing import Literal, Optional, Union

import pytest
from pydantic import BaseModel, Field, ValidationError

from dvcx.lib.dataset import C
from dvcx.lib.feature import (
    Feature,
    ShallowFeature,
)
from dvcx.lib.feature_types import FeatureTypes, pydantic_to_feature
from dvcx.sql.types import JSON, Array, Binary, Boolean, DateTime, Int, String


class FileBasic(ShallowFeature):
    parent: str = Field(default="")
    name: str
    size: int = Field(default=0)


class FileInfo(FileBasic):
    location: dict = Field(default={})


class FileInfoEx(ShallowFeature):
    f_info: FileInfo
    type_id: int


class MyNestedClass(Feature):
    type: int
    Name: str = Field(default="test1")


class MyTest(Feature):
    ThisIsName: str
    subClass: MyNestedClass  # noqa: N815


def test_flatten_schema():
    schema = FileInfo._to_udf_spec()

    assert len(schema) == 4
    assert [item[0] for item in schema] == ["parent", "name", "size", "location"]
    assert [item[1] for item in schema] == [String, String, Int, JSON]


def test_type_datatype():
    class Test1(Feature):
        d: datetime

    schema = Test1._to_udf_spec()
    assert schema[0][1] == DateTime


def test_type_optional_int():
    class Test1(Feature):
        d: Optional[int] = 23

    schema = Test1._to_udf_spec()
    assert schema[0][1] == Int


def test_type_bytes():
    class Test1(Feature):
        d: bytes

    schema = Test1._to_udf_spec()
    assert schema[0][1] == Binary


def test_type_array():
    class Test1(Feature):
        d: list

    schema = Test1._to_udf_spec()
    assert type(schema[0][1]) == Array


def test_type_arrays():
    class Test1(Feature):
        d1: list[int]
        d2: list[float]

    schema = Test1._to_udf_spec()

    assert schema[0][1].to_dict() == {"item_type": {"type": "Int"}, "type": "Array"}
    assert schema[1][1].to_dict() == {"item_type": {"type": "Float"}, "type": "Array"}


def test_type_array_of_arrays():
    class Test1(Feature):
        d1: list[list[int]]

    schema = Test1._to_udf_spec()

    type1 = schema[0][1]
    assert list == type1.python_type
    assert list == type1.impl.item_type.python_type
    assert int == type1.impl.item_type.impl.item_type.python_type


def test_type_json():
    class Test1(Feature):
        d: dict

    schema = Test1._to_udf_spec()
    assert schema[0][1] == JSON


def test_type_bool():
    class Test1(Feature):
        d: bool

    schema = Test1._to_udf_spec()
    assert schema[0][1] == Boolean


def test_type_typed_json():
    class Test1(Feature):
        d: Optional[dict[str, int]]

    schema = Test1._to_udf_spec()
    assert schema[0][1] == JSON


def test_unknown_type():
    class Test1(Feature):
        d: Optional[Decimal]

    with pytest.raises(RuntimeError):
        Test1._to_udf_spec()


def test_flatten_nested_schema():
    schema = FileInfoEx._to_udf_spec()

    assert len(schema) == 5
    assert [item[0] for item in schema] == [
        "parent",
        "name",
        "size",
        "location",
        "type_id",
    ]
    assert [item[1] for item in schema] == [String, String, Int, JSON, Int]


def test_flatten_nested_schema_shallow():
    class MyTest1(Feature):
        a: int = Field(default=33)

    class MyTest2(Feature):
        next2: MyTest1

    class MyTest3(Feature):
        next3: MyTest2

    schema = MyTest3(next3=MyTest2(next2=MyTest1()))._to_udf_spec()
    assert [item[0] for item in schema] == ["my_test3__next3__next2__a"]

    class MyTest3Shallow(ShallowFeature):
        next3: MyTest2

    shallow_schema = MyTest3Shallow(next3=MyTest2(next2=MyTest1()))._to_udf_spec()
    assert [item[0] for item in shallow_schema] == ["a"]


def test_flatten_schema_list():
    t1 = FileInfo(name="test1")
    t2 = FileInfo(name="t2", parent="pp1")
    res = Feature._features_to_udf_spec([t1, t2])
    assert len(t1.model_dump()) == len(res)


def test_flatten_basic():
    vals = FileBasic(parent="hello", name="world", size=123)._flatten()
    assert vals == ("hello", "world", 123)


def test_flatten_with_json():
    t1 = FileInfo(parent="prt4", name="test1", size=42, location={"ee": "rr"})
    assert t1._flatten() == ("prt4", "test1", 42, {"ee": "rr"})


def test_flatten_with_empty_json():
    with pytest.raises(ValidationError):
        FileInfo(parent="prt4", name="test1", size=42, location=None)


def test_flatten_with_accepted_empty_json():
    class Test1(Feature):
        d: Optional[dict]

    assert Test1(d=None)._flatten() == (None,)


def test_flatten_nested():
    t0 = FileInfo(parent="sfo", name="sf", size=567, location={"42": 999})
    t1 = FileInfoEx(f_info=t0, type_id=1849)

    assert t1._flatten() == ("sfo", "sf", 567, {"42": 999}, 1849)


def test_flatten_list():
    t1 = FileInfo(parent="p1", name="n4", size=3, location={"a": "b"})
    t2 = FileInfo(parent="p2", name="n5", size=2, location={"c": "d"})

    vals = t1._flatten_list([t1, t2])
    assert vals == ("p1", "n4", 3, {"a": "b"}, "p2", "n5", 2, {"c": "d"})


@pytest.mark.parametrize("feature_class", [Feature, ShallowFeature])
def test_inheritance(feature_class):
    class SubObject(feature_class):
        subname: str

    class MyTest(feature_class):
        name: str
        sub: SubObject

    class MyTest2(MyTest):
        pass

    with pytest.raises(ValueError):
        MyTest2()

    obj = MyTest2(name="name", sub=SubObject(subname="subname"))
    assert obj._flatten() == ("name", "subname")


def test_naming_transform():
    assert [name for name, _ in MyTest._to_udf_spec()] == [
        "my_test__this_is_name",
        "my_test__sub_class__type",
        "my_test__sub_class__name",
    ]


def test_capital_letter_naming():
    class CAPLetterTEST(Feature):
        AAA_id: str

    vals = [name for name, _ in CAPLetterTEST._to_udf_spec()]
    assert vals == ["cap_letter_test__aaa_id"]


def test_delimiter_in_name():
    with pytest.raises(RuntimeError):

        class MyClass(Feature):
            var__name: str


def test_delimiter_in_name_is_allowed_for_shallow_class():
    class MyClass(ShallowFeature):
        var__name: str

    MyClass._to_udf_spec()

    obj = MyClass(var__name="some stuff")
    dump = obj.model_dump()
    obj._unflatten(dump)


def test_custom_delimiter_in_name():
    with pytest.raises(RuntimeError):

        class MyClass(Feature):
            _delimiter = "EE"
            is_ieee_member: bool


def test_naming_delimiter():
    class MyNestedClassNew(MyNestedClass):
        _delimiter = "+++"

    class MyTestNew(MyTest):
        _delimiter = "+++"

        ThisIsName: str
        subClass: MyNestedClassNew  # noqa: N815

    schema = MyTestNew._to_udf_spec()

    assert [name for name, _ in schema] == [
        "my_test_new+++this_is_name",
        "my_test_new+++sub_class+++type",
        "my_test_new+++sub_class+++name",
    ]


def test_deserialize_nested():
    class Child(Feature):
        type: int
        name: str = Field(default="test1")

    class Parent(Feature):
        name: str
        child: Child

    in_db_map = {
        "parent__name": "a1",
        "parent__child__type": 42,
        "parent__child__name": "a2",
    }

    p = Parent._unflatten(in_db_map)

    assert p.name == "a1"
    assert p.child.type == 42
    assert p.child.name == "a2"


def test_deserialize_shallow():
    class Child(ShallowFeature):
        type: int
        child_name: str = Field(default="test1")

    class Parent(ShallowFeature):
        name: str
        child: Child

    in_db_map = {
        "child_name": "a1",
        "type": 42,
        "name": "a2",
    }

    p = Parent._unflatten(in_db_map)

    assert p.name == "a2"
    assert p.child.type == 42
    assert p.child.child_name == "a1"


def test_deserialize_nested_with_name_normalization():
    class ChildClass(Feature):
        type: int
        name: str = Field(default="test1")

    class Parent(Feature):
        name: str
        childClass11: ChildClass  # noqa: N815

    in_db_map = {
        "parent__name": "name1",
        "parent__child_class11__type": 12,
        "parent__child_class11__name": "n2",
    }

    p = Parent._unflatten(in_db_map)

    assert p.name == "name1"
    assert p.childClass11.type == 12
    assert p.childClass11.name == "n2"


def test_type_array_of_floats():
    class Test1(Feature):
        d: list[float]

    dict_ = {"d": [1, 3, 5]}
    t = Test1(**dict_)
    assert t.d == [1, 3, 5]


def test_class_attr_resolver_basic():
    class MyTest6(Feature):
        val1: list[float]
        pp: int

    assert MyTest6.val1.name == "my_test6__val1"
    assert MyTest6.pp.name == "my_test6__pp"
    assert type(MyTest6.pp.type) == Int
    assert type(MyTest6.val1.type) == Array


def test_class_attr_resolver_shallow():
    class MyTest6(ShallowFeature):
        val1: list[float]
        pp: int

    assert MyTest6.val1.name == "val1"
    assert MyTest6.pp.name == "pp"
    assert type(MyTest6.pp.type) == Int
    assert type(MyTest6.val1.type) == Array


def test_class_attr_resolver_nested():
    assert MyTest.subClass.type.name == "my_test__sub_class__type"
    assert MyTest.subClass.Name.name == "my_test__sub_class__name"
    assert type(MyTest.subClass.type.type) == Int
    assert type(MyTest.subClass.Name.type) == String


def test_class_attr_resolver_nested_3levels():
    class MyTest1(Feature):
        a: int

    class MyTest2(Feature):
        b: MyTest1

    class MyTest3(Feature):
        c: MyTest2

    assert MyTest3.c.b.a.name == "my_test3__c__b__a"
    assert type(MyTest3.c.b.a.type) == Int


def test_list_of_dicts_as_dict():
    class MyTest1(Feature):
        val1: Union[dict, list[dict]]
        val2: Optional[Union[list[dict], dict]]

    schema = Feature._features_to_udf_spec([MyTest1])
    assert len(schema) == 2
    assert next(iter(schema.values())) == JSON
    assert list(schema.values())[1] == JSON


def test_create_column():
    name = "my_col"
    default = "NuLL"

    fr_cls = FeatureTypes.column_class(name, str, default)

    assert issubclass(fr_cls, Feature)
    assert isinstance(fr_cls.my_col, C)
    assert fr_cls.my_col.name == name

    val = "1qa"
    assert fr_cls(my_col=val).my_col == val
    assert fr_cls().my_col == default


def test_create_columns():
    name1 = "cl11"
    name2 = "m234q"

    def1 = "d1wer"
    def2 = 7654

    fr_cls = FeatureTypes.column_classes({name1: (str, def1), name2: (int, def2)})

    assert issubclass(fr_cls, Feature)
    assert isinstance(fr_cls.cl11, C)
    assert isinstance(fr_cls.m234q, C)

    val1 = "deui"
    val2 = 126
    fr = fr_cls(cl11=val1, m234q=val2)
    assert fr.cl11 == val1
    assert fr.m234q == val2

    assert fr_cls().cl11 == def1
    assert fr_cls().m234q == def2


def test_get_value_with_check():
    my_prefix = "111"
    my_suffix = "333"
    def_suffix = "ddd"
    my_value = "22"

    class MyColumn(FeatureTypes.column_class("val", str)):
        def get_value(self, prefix: str, suffix: str = def_suffix) -> str:
            return prefix + super().get_value() + suffix

    fr = MyColumn(val=my_value)

    assert (
        fr._get_value_with_check(my_prefix, my_suffix)
        == my_prefix + my_value + my_suffix
    )
    assert fr._get_value_with_check(my_prefix) == my_prefix + my_value + def_suffix

    with pytest.raises(ValueError):
        fr._get_value_with_check(suffix=my_suffix)

    with pytest.raises(ValueError):
        fr._get_value_with_check()

    assert (
        fr._get_value_with_check(prefix=my_prefix) == my_prefix + my_value + def_suffix
    )


def test_column_class_basic():
    cls = FeatureTypes.column_class("field_name5")
    for type_, value in [
        (str, "val13"),
        (int, 873),
        (dict, {"x": 123, "y": 23.32}),
        (list, ["ee", 34.4, 432]),
    ]:
        obj = cls(field_name5=value)
        assert obj.field_name5 == value
        assert type(obj.field_name5) is type_

    cls_int = FeatureTypes.column_class("field33", int)
    with pytest.raises(ValidationError):
        cls_int(field33="value")

    cls_str = FeatureTypes.column_class("field_76", str)
    with pytest.raises(ValidationError):
        cls_str(field_76=3.14)


def test_column_class_from_sql_column():
    field = "field11"

    for column, value in [
        (C(field, str), "val123"),
        (C(field, int), 463),
        (C(field, float), 1.62),
        (C(field, list), ["a", "b", 33]),
        # No type:
        (C(field), "val123"),
        (C(field), 463),
        (C(field), 1.62),
        (C(field), ["a", "b", 33]),
    ]:
        cls = FeatureTypes.column_class(column)
        obj = cls(field11=value)
        assert obj.field11 == value

    cls_int = FeatureTypes.column_class(C(field, int))
    assert cls_int(field11=123) is not None
    with pytest.raises(ValidationError):
        cls_int(field11="sddsd")
    with pytest.raises(ValidationError):
        cls_int(field11={"ee": "hey"})

    cls_str = FeatureTypes.column_class(C(field, str))
    with pytest.raises(ValidationError):
        cls_str(field11=52)


def test_literal():
    class MyTextBlock(Feature):
        id: int
        type: Literal["text"]

    schema = Feature._features_to_udf_spec([MyTextBlock])
    assert len(schema) == 2
    assert next(iter(schema.values())) == Int
    assert list(schema.values())[1] == String


def test_pydantic_to_feature():
    class MyTextBlock(BaseModel):
        id: int
        type: Literal["text"]

    cls = pydantic_to_feature(MyTextBlock)
    assert Feature._is_feature_class(cls)

    schema = Feature._features_to_udf_spec([cls])
    assert len(schema) == 2
    assert next(iter(schema.values())) == Int
    assert list(schema.values())[1] == String


def test_pydantic_to_feature_nested():
    class MyTextBlock(BaseModel):
        id: int
        type: Literal["text"]

    class MyMessage(BaseModel):
        val1: Optional[str]
        val2: MyTextBlock

    cls = pydantic_to_feature(MyMessage)
    assert Feature._is_feature_class(cls)

    schema = Feature._features_to_udf_spec([cls])
    assert len(schema) == 3
    assert next(iter(schema.values())) == String
    assert list(schema.values())[1] == Int
    assert list(schema.values())[2] == String


def test_flatten_schema_with_list_of_objects():
    class FileDir(Feature):
        name: str
        files: list[FileBasic]

    schema = FileDir._to_udf_spec()

    assert len(schema) == 2
    assert [item[0] for item in schema] == ["file_dir__name", "file_dir__files"]

    assert schema[0][1] == String
    assert schema[1][1].to_dict() == {"item_type": {"type": "JSON"}, "type": "Array"}


def test_flatten_schema_with_list_of_ints():
    class SomeInfo(Feature):
        name: str
        vals: list[int]

    schema = SomeInfo._to_udf_spec()

    assert len(schema) == 2
    assert [item[0] for item in schema] == ["some_info__name", "some_info__vals"]

    assert schema[0][1] == String
    assert schema[1][1].to_dict() == {"item_type": {"type": "Int"}, "type": "Array"}



class DVCXError(RuntimeError):
    pass


class NotFoundError(Exception):
    pass


class DatasetNotFoundError(NotFoundError):
    pass


class DatasetInvalidVersionError(Exception):
    pass


class StorageNotFoundError(NotFoundError):
    pass


class PendingIndexingError(Exception):
    """An indexing operation is already in progress."""


class QueryScriptCompileError(Exception):
    pass


class QueryScriptRunError(Exception):
    """Error raised by `subprocess.run`.

    Attributes:
        message      Explanation of the error
        return_code  Code returned by the subprocess
        output       STDOUT + STDERR output of the subprocess
    """

    def __init__(self, message: str, return_code: int = 0, output: str = ""):
        self.message = message
        self.return_code = return_code
        self.output = output
        super().__init__(self.message)


class QueryScriptDatasetNotFound(QueryScriptRunError):  # noqa: N818
    pass


class QueryScriptCancelError(QueryScriptRunError):
    pass


class ClientError(RuntimeError):
    def __init__(self, message, error_code=None):
        super().__init__(message)
        # error code from the cloud itself
        self.error_code = error_code


class TableMissingError(DVCXError):
    pass



import os
from collections.abc import Mapping
from typing import TYPE_CHECKING, Optional

from tomlkit import load

if TYPE_CHECKING:
    from tomlkit import TOMLDocument


def read_config(dvcx_root: str) -> Optional["TOMLDocument"]:
    config_path = os.path.join(dvcx_root, "config")
    try:
        with open(config_path, encoding="utf-8") as f:
            return load(f)
    except FileNotFoundError:
        return None


def get_remote_config(
    config: Optional["TOMLDocument"], remote: str = ""
) -> Mapping[str, str]:
    if config is None:
        return {"type": "local"}
    if not remote:
        try:
            remote = config["core"]["default-remote"]  # type: ignore[index,assignment]
        except KeyError:
            return {"type": "local"}
    try:
        remote_conf: Mapping[str, str] = config["remote"][remote]  # type: ignore[assignment,index]
    except KeyError:
        raise Exception(
            f"missing config section for default remote: remote.{remote}"
        ) from None
    except Exception as exc:
        raise Exception("invalid config") from exc

    if not isinstance(remote_conf, Mapping):
        raise TypeError(f"config section remote.{remote} must be a mapping")

    remote_type = remote_conf.get("type")
    if remote_type not in ("local", "http"):
        raise Exception(
            f'config section remote.{remote} must have "type" with one of: '
            '"local", "http"'
        )

    if remote_type == "http":
        for key in ["url", "username", "token"]:
            try:
                remote_conf[key]
            except KeyError:
                raise Exception(
                    f"config section remote.{remote} of type {remote_type} "
                    f"must contain key {key}"
                ) from None
    elif remote_type != "local":
        raise Exception(
            f"config section remote.{remote} has invalid remote type {remote_type}"
        )
    return remote_conf



from typing import Optional

from pydantic import Field

from dvcx.lib.webdataset import WDSBasic, WDSReadableSubclass


class Laion(WDSReadableSubclass):
    uid: str = Field(default="")
    face_bboxes: Optional[list[list[float]]] = Field(default=None)
    caption: Optional[str] = Field(default=None)
    url: Optional[str] = Field(default=None)
    key: Optional[str] = Field(default=None)
    status: Optional[str] = Field(default=None)
    error_message: Optional[str] = Field(default=None)
    width: Optional[int] = Field(default=None)
    height: Optional[int] = Field(default=None)
    original_width: Optional[int] = Field(default=None)
    original_height: Optional[int] = Field(default=None)
    exif: Optional[str] = Field(default=None)
    sha256: Optional[str] = Field(default=None)

    @staticmethod
    def _reader(builder, item):
        return Laion.model_validate_json(builder.read_text(item))


class WDSLaion(WDSBasic):
    txt: Optional[str] = Field(default=None)
    json: Laion  # type: ignore[assignment]



from sqlalchemy.sql.functions import GenericFunction

from dvcx.sql.types import Float, Int64
from dvcx.sql.utils import compiler_not_implemented


class cosine_distance(GenericFunction):  # noqa: N801
    type = Float()
    package = "array"
    name = "cosine_distance"
    inherit_cache = True


class euclidean_distance(GenericFunction):  # noqa: N801
    type = Float()
    package = "array"
    name = "euclidean_distance"
    inherit_cache = True


class length(GenericFunction):  # noqa: N801
    type = Int64()
    package = "array"
    name = "length"
    inherit_cache = True


class sip_hash_64(GenericFunction):  # noqa: N801
    type = Int64()
    package = "hash"
    name = "sip_hash_64"
    inherit_cache = True


compiler_not_implemented(cosine_distance)
compiler_not_implemented(euclidean_distance)
compiler_not_implemented(length)
compiler_not_implemented(sip_hash_64)
